#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "udmsearch.h"

static int seconds	=0; /* To sleep between documents	*/
static int flags	=0; /* For indexer			*/
static int started	=0; /* If initialization already done	*/
static int total_threads=0; /* Total threads number		*/
static int sleep_threads=0; /* Number of sleepping threads      */
static int next_thread  =1; /* Handle number for indexer	*/
static char *config_name= UDM_CONF_DIR "/indexer.conf";

#if (WIN32|WINNT)
#ifdef HAVE_PTHREAD
static CRITICAL_SECTION main_mutex;
#endif
#else
#include <unistd.h>
#ifdef HAVE_PTHREAD
static pthread_mutex_t main_mutex;
#endif
#endif



int usage(){

	fprintf(stderr,
"\n\
indexer from UdmSearch v." VERSION "/%s\n\
http://mysearch.udm.net (C) 1998-2000, UdmSearch Developers Team.\n\
\n\
Usage: indexer [OPTIONS]  [configfile]\n\
\n\
Indexing options:\
"
#ifdef HAVE_SQL
"\n\
  -a            reindex all documents even if not expired (may be\n\
                limited using -t, -u, -s and -f options)\n\
  -m            reindex expired documents even if not modified (may\n\
                be limited using -t, -u, and -s options)\n\
  -e            index 'most expired' (oldest) documents first\n\
  -o            index documents with less depth (hops value) first\n\
  -n n          index only n documents and exit\n\
  -q            quick startup (do not add Server URLs)\n\
  -k            skip locking (affects for MySQL and PostgreSQL)\n\
"
#endif
"\n\
  -i            insert new URLs (URLs to insert must be given using -u or -f)\n\
  -p n          sleep n seconds after each URL\n\
  -w            do not warn before clearing documents from database\n\
"
#ifdef HAVE_PTHREAD
"  -N n         run N threads\n\
"
#endif

#ifdef HAVE_SQL
"\n\
Subsection control options (may be combined):\n\
  -s status     limit indexer to documents matching status (HTTP Status code)\n\
  -t tag        limit indexer to documents matching tag\n\
  -u pattern    limit indexer to documents with URLs matching pattern\n\
                (supports SQL LIKE wildcard '%%')\n\
  -f filename   read URLs to be indexed/inserted/cleared from file (with -a\n\
                or -C option, supports SQL LIKE wildcard '%%'; has no effect\n\
                when combined with -m option)\n\
  -f -          Use STDIN instead of file as URL list\n\
"
#else
"\n\
URL options:\n\
  -u URL        insert URL at startup\n\
  -f filename   read URLs to be inserted from file\n\
"
#endif
"\n\
Logging options:\n\
"
#ifdef LOG_PERROR
"  -l            do not log to stdout/stderr\n\
"
#endif
"  -v n          verbose level, 0-5\n\
\n\
Misc. options:\n\
"
#ifdef HAVE_SQL
"\n\
  -C            clear database and exit\n\
  -S            print statistics and exit\n\
  -I            print referers and exit\n\
"
#endif
"  -h,-?         print this help page and exit\n\
\n\
\n\
Mail bug reports and suggestions to <udmsearch@search.udm.net>.\n",DBVersion);

	return(0);
}



static int IndexerMain(UDM_INDEXER * Indexer){
int res=IND_OK;
int i_sleep=0;

	while(1){
		
		LOCK_MAIN();
		if(!started){
			res=IndexNextURL(Indexer,flags|UDM_FLAG_INIT);
			started=1;
		}
		UNLOCK_MAIN();

		if(res!=IND_ERROR) /* Possible after bad startup */
			res=IndexNextURL(Indexer,flags);

		switch(res){
			case IND_OK:
				if(i_sleep){
					LOCK_MAIN();
					sleep_threads--;
					UNLOCK_MAIN();
					i_sleep=0;
				}
				continue;

			case IND_ERROR:
				THREADINFO(Indexer->handle,"Error",IndexerErrorMsg(Indexer));
				udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				return(1);

			case IND_NO_TARGET:
#ifdef HAVE_PTHREAD
			/* in multi-threaded environment we		*/
			/* should wait for a moment when every thread	*/
			/* has nothing to do				*/

				if(!i_sleep){
					LOCK_MAIN();
					sleep_threads++;
					UNLOCK_MAIN();
					i_sleep=1;
				}

				LOCK_MAIN();
				res=(sleep_threads>=total_threads);
				UNLOCK_MAIN();

				if(res)return(0);
				continue;
#else
				return(0);
#endif
			default:
				return(0);
		}
		if(seconds){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"Sleeping %d second(s)",seconds);
			UDMSLEEP(seconds);
		}
	}
	return(0);
}


#if  (WIN32|WINNT)
DWORD WINAPI thread_main(void *arg){
#else
void * thread_main(void *arg){
#endif
UDM_INDEXER * Indexer;

	LOCK_MAIN();
	Indexer=AllocIndexer(next_thread++);
	UNLOCK_MAIN();

	IndexerMain(Indexer);

	LOCK_MAIN();
	udm_log(Indexer->handle,UDM_LOG_ERROR,"Done");
	total_threads--;
	UNLOCK_MAIN();

	FreeIndexer(Indexer);

#if     (WIN32|WINNT)
	return(0);
#else
	return(NULL);
#endif
}


int main(int argc, char **argv) {

char *language=NULL;
int npages=-1;
int clear=0,stat=0,integrity=0;
int log2stderr=1,dump=0;
int add_servers=1;
extern char *optarg;
extern int optind;
int ch;
int maxthreads=1;
int warnings=1;
char * url_fname=NULL;

	while ((ch = getopt(argc, argv, "CSIaheolmdqiwk?t:u:s:n:v:L:A:D:p:N:f:")) != -1){
		switch (ch) {
		case 'C': clear++;add_servers=0;break;	
		case 'S': stat++;add_servers=0;break;
		case 'I': integrity++;add_servers=0;break;
		case 'L': language=optarg;break;
		case 'q': add_servers=0;break;
		case 'l': log2stderr=0;break;
		case 'a': flags|=UDM_FLAG_MARK;break;
		case 'e': flags|=UDM_FLAG_EXP_FIRST;break;
		case 'o': flags|=UDM_FLAG_SORT_HOPS;break;
		case 'm': flags|=UDM_FLAG_REINDEX;break;
		case 'k': flags|=UDM_FLAG_SKIP_LOCKING;break;
		case 'n': npages=atoi(optarg);break;
		case 'v': SetLogLevel(atoi(optarg));break;
		case 'p': seconds=atoi(optarg);break;
		case 'd': dump=1;break;
		case 't': AddTagLimit(atoi(optarg));break;
		case 's': AddStatusLimit(atoi(optarg));break;
		case 'u': 
			AddURLLimit(optarg);
			if(flags&UDM_FLAG_INSERT)
				AddHref(NULL,optarg,0,0,0);
			break;
		case 'N': maxthreads=atoi(optarg);break;
		case 'f': url_fname=optarg;break;
		case 'i': flags|=UDM_FLAG_INSERT;break;
		case 'w': warnings=0;break;
		case '?':
		case 'h':
		default:  
			usage();
			return(1);
		}
	}

	argc -= optind;argv += optind;

	if(argc>1){
		usage();
		return(1);
	}
	if(argc==1)config_name=argv[0];

	InitCharset();

#ifdef HAVE_PTHREAD
	InitMutexes();
#if (WIN32|WINNT)
	InitializeCriticalSection(&main_mutex);
#else
	pthread_mutex_init(&main_mutex,NULL);
#endif
#endif


	if(LoadConfig(config_name,0,
		(add_servers?UDM_FLAG_ADD_SERV:0)+
		((!integrity&&!stat&&!clear)*UDM_FLAG_SPELL))){		
			fprintf(stderr,"%s\n",ConfErrMsg());
			exit(1);
	}

#if (WIN32|WINNT)
#else
	udm_openlog(log2stderr);
#endif


	/* Make sure URL file is readable if not STDIN */
	if(url_fname) {
		AddURLFile(url_fname);
		if(strcmp(url_fname,"-")){
			FILE *url_file;
			if(!(url_file=fopen(url_fname,"r"))){
				udm_log(0,UDM_LOG_ERROR,"Error: can't open url file '%s': %s",url_fname, strerror(errno));
				exit(1);
			}
			fclose(url_file);
		}
	}

	/* Make sure all URLs to be inserted are OK */
	if(flags&UDM_FLAG_INSERT) {
		/*
		if(uurl) {
			URL myurl;
			if(parseurl(&myurl,uurl)) {
				udm_log(0,UDM_LOG_ERROR,"Error: Invalid URL '%s'",uurl);
				exit(1);
			}
		}
		*/

		if(url_fname) {
			if(strcmp(url_fname,"-")){
				if(IND_OK!=URLFile(NULL, UDM_URL_FILE_PARSE)){
					udm_log(0,UDM_LOG_ERROR,"Error: Invalid URL in '%s'",url_fname);
					exit(1);
				}
			}
		}
	}

	if(clear){
		UDM_INDEXER * Indexer;
		int clear_confirmed=0;
		if(warnings) {
			char str[5]="";
			printf("You are going to delete database '%s' content\n",DBName?DBName:"");
			printf("Are you sure?(YES/no)");
			if(fgets(str,sizeof(str),stdin)){
				if(strncmp(str,"YES",3))
					printf("Canceled\n");
				else
					clear_confirmed=1;
			}
		}
		else
			clear_confirmed=1;

		if(clear_confirmed) {
			Indexer=AllocIndexer(0);
			if(url_fname) {
				if(IND_OK!=URLFile(Indexer,UDM_URL_FILE_CLEAR)){
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				}
			}
			else {
				if(IND_OK!=ClearDatabase(Indexer)){
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				}
			}
			FreeIndexer(Indexer);
		}else{
			printf("Canceled\n");
		}
	}else
	if(stat){
		UDM_INDEXER * Indexer;
		Indexer=AllocIndexer(0);
		if(IND_OK!=ShowStatistics(Indexer)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
		}
		FreeIndexer(Indexer);
	}else
	if(integrity){
		UDM_INDEXER * Indexer;
		Indexer=AllocIndexer(0);
		if(IND_OK!=ShowReferers(Indexer)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
		}
		FreeIndexer(Indexer);
	}else{
#if (WIN32|WINNT)
#else
		udm_log(0,UDM_LOG_ERROR, "indexer from UdmSearch v." VERSION"/%s started with '%s'", DBVersion, config_name);
#endif
		SetMaxURLNumber(npages);

#ifdef HAVE_PTHREAD
		{
#if (WIN32|WINNT)
#else
			pthread_t *thread;
#endif
			int i;
			total_threads=maxthreads;
			LOCK_MAIN();
			for(i=0;i<maxthreads;i++){
#if (WIN32|WINNT)
				CreateThread(NULL, 0, &thread_main, NULL, 0, NULL);
#else
				thread=(pthread_t*)malloc(sizeof(pthread_t));
				pthread_create(thread,NULL,&thread_main,NULL);
#endif
				
			}
			UNLOCK_MAIN();
		}

		while(1){
			int num;
			LOCK_MAIN();
			num=total_threads;
			UNLOCK_MAIN();
			if(!num)break;
		}
		DestroyMutexes();

#if (WIN32|WINNT)
		DeleteCriticalSection(&main_mutex);
#else
		pthread_mutex_destroy(&main_mutex);
#endif
#else
		thread_main(NULL);
#endif
	}
	total_threads=0;

	return(0);
}
